/***************************************************************************
                          sqlguidoc_mysql.cpp
                          -------------------
    begin                : 08/02 2002
    copyright            : (C) 2002 by mb
    email                : marcus@beranek.de
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


/*
  This class implements all necessary member-functions of sqlguidoc, which
  must be reimplenented for a special database (here MySQL).
*/

// include files for Qt
#include <qdir.h>
#include <qwidget.h>
#include <qstring.h>
#include <qapplication.h>
#include <qregexp.h>

// include files for KDE
#include <kapplication.h>
#include <klocale.h>
#include <kmessagebox.h>
#include <kio/job.h>

// application specific includes
#include "sqlguidoc_pgsql.h"

#include <qcursor.h>
#include <qapplication.h>

#include "dbtypes.h"

#include <iostream>
#include <iomanip>
using namespace std;


SqlGuiDocPgsql::SqlGuiDocPgsql(QWidget *parent, const char *name) : SqlGuiDoc(parent, name){
  // nothing to do.
}

SqlGuiDocPgsql::~SqlGuiDocPgsql(){

}

QString SqlGuiDocPgsql::getProtocol(){
  QString p = "pgsql";
  return p;
}

void SqlGuiDocPgsql::slotDbCreate( QString* db ){
  
  if(db->isEmpty()) return;

  QString * cr  = new QString("CREATE DATABASE ");
  QString * def = new QString(""); // databasename empty, so we use the default-database to connect to...
  cr->append( *db );

  slotDoQuery( def, cr);
  emit( NewDbCreated() );

}


void SqlGuiDocPgsql::slotCountRows( QString* _db, QString* _tb){
  QApplication::setOverrideCursor( Qt::waitCursor );

  // std::cout << "Got Count-Row-Request..." << endl;

  if(_db->isEmpty()){
    QApplication::restoreOverrideCursor();
    return;
  }
  if(_tb->isEmpty()){
    QApplication::restoreOverrideCursor();
    return;
  }

  QString * q = new QString("SELECT count(*) AS ROWS FROM ");
  q->append( *_tb );

  slotDoQuery( _db, q);

  QApplication::restoreOverrideCursor();

}

int SqlGuiDocPgsql::getDBType(){
  return SQLGUI_PGSQL;
}

void SqlGuiDocPgsql::slotDbDelete( QString* db ){

  if(db->isEmpty()) return;

  QString * cr  = new QString("DROP DATABASE ");
  QString * def = new QString(""); // databasename empty, so we use the default-database to connect to...
  cr->append( *db );

  slotDoQuery( def, cr);
  emit( DbErased() );

}

void SqlGuiDocPgsql::slotTableDelete( QString* db, QString* tb ){

  if(db->isEmpty()) return;
  if(tb->isEmpty()) return;
  
  QString * q = new QString("DROP TABLE ");
  q->append( *tb );
  
  slotDoQuery( db, q );
  emit( TableErased() );

}


void SqlGuiDocPgsql::slotServerStatus( ){

  // std::cout << "Got Staus-Request..." << endl;

  QString *status = new QString();

  QString *pNum = new QString("");
  if(doc_port>0)pNum->setNum( doc_port );
  
  QString * cinfo = new QString("host='");
  cinfo->append( doc_host.latin1() );
  cinfo->append("' user='");
  cinfo->append( doc_uname.latin1() );
  cinfo->append("' port='");
  cinfo->append( pNum->latin1() );
  cinfo->append( "' dbname='" );
  cinfo->append( "' password='" );
  cinfo->append( doc_pass.latin1() );
  cinfo->append( "'" );

  PGconn     *conn;
  PGresult   *res;
  
  conn = PQconnectdb( cinfo->latin1() );
  if ( PQstatus(conn) == CONNECTION_BAD ){
    // std::cout << "no conn! ERROR connecting to server... :-((" << endl;
    // std::cout << PQerrorMessage(conn) << endl;
    QString *errmsg = new QString( kl->translate( i18n("Couldn't get server-status."))) ;
    PQfinish(conn);
    emit( DbQueryError( errmsg) );
    return;

  } else {
    
    res = PQexec(conn, "BEGIN");
    if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
      {
        // std::cout << "no begin! ERROR connecting to server... :-((" << endl;
        PQclear(res);
        PQfinish(conn);
        QString *errmsg = new QString( kl->translate( i18n("Couldn't get server-status."))) ;
        emit( DbQueryError( errmsg) );
        return;
      }
    PQclear(res);
    QString *sql = new QString();
    sql->append("DECLARE mycursor CURSOR FOR ");
    sql->append( "SELECT version(),current_user(),session_user()" );
    
    res = PQexec(conn, *sql);
    if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
      {
        // std::cout << "ERROR connecting to server... :-((" << endl;
        // std::cout << PQerrorMessage(conn) << endl;
        
        PQclear(res);
        PQfinish(conn);
        QString *errmsg = new QString( kl->translate( i18n("Couldn't get server-status."))) ;
        emit( DbQueryError( errmsg) );
        return;
      }

    PQclear(res);
    res = PQexec(conn, "FETCH ALL in mycursor");
    if (!res || PQresultStatus(res) != PGRES_TUPLES_OK)
      {
        // std::cout << "no fetchall! ERROR connecting to server... :-((" << endl;
        PQclear(res);
        PQfinish(conn);
        QString *errmsg = new QString( kl->translate( i18n("Couldn't get server-status."))) ;
        emit( DbQueryError( errmsg) );
        return;
      }
      
    int nFields = PQnfields(res);
    int nRows =  PQntuples(res);
    
    for (int i = 0; i < nRows; i++){
      
      for (int j = 0; j < nFields; j++){
        
        char * stat = new char[ strlen( PQgetvalue(res, i, j) ) + 1 ];
        
        if( strlen( PQgetvalue(res, i, j) ) > 0 )strcpy( stat, PQgetvalue(res, i, j) );
        else stat=NULL;

        QString *title = new QString();
        switch(j){
        case 0: {
          title->append("Version: ");
          break;
        }
        case 1: {
          title->append("Current user: ");
          break;
        }
        case 2: {
          title->append("Session user: ");
          break;
        }
        }
        status->append( *title );
        status->append( stat );
        status->append( "\n" );
        
      }
    }

    PQclear(res);
    res = PQexec(conn, "CLOSE mycursor");
    PQclear(res);
    res = PQexec(conn, "COMMIT");
    PQclear(res);
    PQfinish(conn);

    emit serverStatus( *status );
  }

}


void SqlGuiDocPgsql::slotTableCheck( QString* db, QString* tb ){

  // std::cout << "Got Table-Check-Request..." << endl;

  if(db->isEmpty()){
    return;
  }
  if(tb->isEmpty()){
    return;
  }
  
  QString * q = new QString("ANALYSE ");
  q->append( *tb );
  
  slotDoQuery( db, q);
  
}

void SqlGuiDocPgsql::slotTableOptimize( QString* db, QString* tb ){

  // std::cout << "Got Table-Optimize-Request..." << endl;

  if(db->isEmpty()){
    return;
  }
  if(tb->isEmpty()){
    return;
  }
  
  QString * q = new QString("VACUUM ");
  q->append( *tb );
  
  slotDoQuery( db, q);
  
}

void SqlGuiDocPgsql::slotTableRepair( QString* db, QString* tb ){

  // std::cout << "Got Table-Repair-Request..." << endl;

  if(db->isEmpty()){
    return;
  }
  if(tb->isEmpty()){
    return;
  }

  QString * q = new QString("REINDEX TABLE ");
  q->append( *tb );

  slotDoQuery( db, q);

}

void SqlGuiDocPgsql::slotDoQuery(QString * Base, QString* Anfrage){

  QString *pNum = new QString("");
  if(doc_port>0)pNum->setNum( doc_port );
  
  QString * cinfo = new QString("host='");
  cinfo->append( doc_host.latin1() );
  cinfo->append("' user='");
  cinfo->append( doc_uname.latin1() );
  cinfo->append("' port='");
  cinfo->append( pNum->latin1() );
  cinfo->append( "' dbname='" );
  cinfo->append( Base->latin1() );
  cinfo->append( "' password='" );
  cinfo->append( doc_pass.latin1() );
  cinfo->append( "'" );
  
  // // std::cout << cinfo->latin1() << endl;

  PGconn     *conn;
  PGresult   *res;

  conn = PQconnectdb( cinfo->latin1() );
  if ( PQstatus(conn) == CONNECTION_BAD ){
    // std::cout << "no conn! ERROR connecting to server... :-((" << endl;
    // std::cout << PQerrorMessage(conn) << endl;
    QString *errmsg = new QString( kl->translate( i18n("Error connecting to server.")));
    QApplication::restoreOverrideCursor();
    
    PQfinish(conn);
    
    emit( DbQueryError( errmsg) );
    QApplication::restoreOverrideCursor();
    return;
  } else {
    
    clearResult();
    base_t * hilf=NULL;
    
    /*
     * Analyse statement, look for SELECT- or CREATE-command.
     */
    bool withTransaction=true;
    bool declareCursor=false;
    if( Anfrage->contains( "SELECT",false ) ){
      declareCursor = true;
    }
    if(   ( Anfrage->contains( "CREATE DATABASE",false ) )   ||   ( Anfrage->contains( "DROP DATABASE",false ) )   ||   ( Anfrage->contains( "ANALYSE",false ) )   ||   ( Anfrage->contains( "VACUUM",false ) )   ||   ( Anfrage->contains( "REINDEX",false ) )   ||   ( Anfrage->contains( "BEGIN",false ) )   ||   ( Anfrage->contains( "END",false ) )   ||   ( Anfrage->contains( "COMMIT",false ) )   ||   ( Anfrage->contains( "ROLLBACK",false ) )   ){
      withTransaction = false;
      declareCursor = false;
    }
    if(withTransaction){
      /* start a transaction block */
      res = PQexec(conn, "BEGIN");
      if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
        {
          // std::cout << "no begin! ERROR connecting to server... :-((" << endl;
          PQclear(res);
          PQfinish(conn);
          QString *errmsg = new QString( kl->translate( i18n("Error connecting to server (No TRANSACTION-BLOCK).")));
          emit( DbQueryError( errmsg) );
          QApplication::restoreOverrideCursor();
          return;
        }
      PQclear(res);
    }
    

    QString *sql = new QString();
    if (declareCursor) { sql->append("DECLARE mycursor CURSOR FOR "); }
    sql->append( *Anfrage );
    
    res = PQexec(conn, *sql);
    if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
      {
        // fprintf(stderr, "DECLARE CURSOR command failed\n");
        // std::cout << "ERROR connecting to server... :-((" << endl;
        // std::cout << PQerrorMessage(conn) << endl;
        
        PQclear(res);
        PQfinish(conn);
        QString *errmsg = new QString( kl->translate( i18n("Error connecting to server (No DECLARE CURSOR / Problems sending query).")));
        emit( DbQueryError( errmsg) );
        QApplication::restoreOverrideCursor();
        return;
      }

    
    if(declareCursor){
      PQclear(res);
      res = PQexec(conn, "FETCH ALL in mycursor");
      if (!res || PQresultStatus(res) != PGRES_TUPLES_OK)
        {
          //fprintf(stderr, "FETCH ALL command didn't return tuples properly\n");
          // std::cout << "no fetchall! ERROR connecting to server... :-((" << endl;
          PQclear(res);
          PQfinish(conn);
          QString *errmsg = new QString( kl->translate( i18n("Error connection to server (No FETCH ALL IN CURSOR).")));
          emit( DbQueryError( errmsg) );
          QApplication::restoreOverrideCursor();
          return;
        }
      
      
      //Spaltennamen setzen:
      int nFields = PQnfields(res);
      column_t * columne;
      for (int i = 0; i < nFields; i++){
        
        //  // std::cout <<  PQfname(res, i) << endl;
        columne = new column_t;
        column_t * column_alt;
        int p = strlen( PQfname(res, i)  );
        char * sp_name = new char[ p+1 ];
        
        if( p>0 ){
          sp_name = strcpy( sp_name, PQfname(res, i) );
        }else{ 
          sp_name=NULL;
        }
        
        columne->name=sp_name;
        unsigned int typus = (unsigned int)  PQftype(res,i);
        QString * tp = new QString();
        tp->setNum( typus );
        /*****************************************************************************************************************/
        // Hier Tabellen-Typ ermitteln, durch Vergleich mit columnType...

        column_t * c_list = columnTypes;
        bool stop = false;
        char* typ = NULL;

        while(!stop){
          if( tp->compare( c_list->type ) == 0 ){
            // Oid gefunden...
            typ = new char[ strlen( c_list->name) + 1 ];
            typ = strcpy( typ, c_list->name );
            stop = true;
          }
          if(c_list){
            c_list = c_list->next;
          } else {
            stop = true;
          }
        }
        columne->type = typ;

        /*****************************************************************************************************************/
       
        if(i==0) lastCols=columne;
        else column_alt->next=columne;
        column_alt=columne;
      }
      if( columne ) columne->next=NULL;
      
      if ( RESULT ){
        delete RESULT;
        RESULT = NULL;
      }
      if ( RESULT_Start ){
        clearResult();
        delete RESULT_Start;
        RESULT_Start = NULL;
      }
      
      row_t * hilf_row = NULL;
      field_t * hilf_col = NULL;    
      
      int nRows =  PQntuples(res);
      
      for (int i = 0; i < nRows; i++){
        
        row_t * rw = new row_t;
        rw->next = NULL;
        for (int j = 0; j < nFields; j++){
          
          char * rw_name = new char[ strlen( PQgetvalue(res, i, j) ) + 1 ];
          field_t * col = new field_t;
          
          if( strlen( PQgetvalue(res, i, j) ) > 0 )strcpy( rw_name, PQgetvalue(res, i, j) );
          else rw_name=NULL;
          
          col->value = rw_name;
          
          if( hilf_col ){
            hilf_col->next = col;
          } else {
            rw->feld = col;
          }
          hilf_col = col;
          
        }
        if(hilf_col)hilf_col->next=NULL;

        if (hilf_row){
          hilf_row->next = rw;
        } else {
          RESULT_Start = rw;
          RESULT = rw;
        }
        
        hilf_row = rw;
        hilf_col = NULL;
      }
      
      if(hilf_row)hilf_row->next=NULL;
      emit( resultChanged() );
                       
      PQclear(res);
      res = PQexec(conn, "CLOSE mycursor");

    } // Ende if declareCursor

    PQclear(res);

    if (withTransaction){
      res = PQexec(conn, "COMMIT");
      PQclear(res);
    }
    PQfinish(conn);
  }

  QString *sql = new QString( Anfrage->simplifyWhiteSpace() );
  QChar   bst = *(new QChar);
  
  lastDB->setLatin1( Base->latin1() );
  
  // std::cout << "Analysing SQL..." << sql->latin1() << endl;
  
  int select = sql->find("select ",0,false);
  
  if (select > -1){
    // Es handelt sich um ein select-statement:
    // std::cout << "Select-Statement..." << endl;
    // wir lschen die alte Variable:
    lastTable->setLatin1("");
    int from_pos = sql->find("from ",0,false);
    if (from_pos > -1){
      // Wir haben einen Tabellen-Namen nach dem From:
      // std::cout << "found from-expression... now looking for table..." << endl;
      from_pos += 5; // das wort "from " hat nochmals 5 buchstaben...
      int l = sql->length();
      bool weiter = true;
      int j = from_pos;
      while ( weiter ){
        bst = sql->at(j);
        lastTable->append( bst );
        j++;
        if(j==l){weiter=false;}
        if(bst == " "){weiter=false;}
      }
      // std::cout << "tablename: " << lastTable->latin1() << endl;
    }
  }
  column_t * sp;
  table_t   * tab;
  base_t    * bse = DB;
  while( bse ){
    //    // std::cout << "Checking DB " << lastDB->latin1() << " - " << bse->name << endl;
    if ( lastDB->compare( bse->name) == 0 ){
      tab = bse->table_list;
      // // std::cout << "Found last db in DB-Object " << endl;
      bse = NULL;
      while( tab ){
        if ( strcmp( tab->name, lastTable->latin1() ) == 0 ){	
          sp = tab->column_list;
          // // std::cout << "Found last table in DB-Object " << endl;
          tab = NULL;
          while(sp){
            if ( sp->pri == 'Y' ){
              // // std::cout << "Found primary-column in last table " << endl;
              primKey->setLatin1( sp->name );
              // // std::cout << primKey->latin1() << endl;
              sp = NULL;
            } else {
              sp = sp->next;
            }
          }
        } else {	
          tab = tab->next;
        }
      }	
    } else {
      bse = bse->next;
    }
  }
  emit lastActionChanged( lastDB, lastTable, primKey);
  emit( QueryDone() );
}

bool SqlGuiDocPgsql::openDocument(const QString& url, const QString& uname, const QString& pwd, const QString& base){
  /*
   * Reading the catalog from postgresql suxx! And the doc is horrible!!!
   * It may take some time, to get the stuff from the database... There may be better ways for getting these information,
   * but I don't have the time and desire to find out...
   *
   */

  const char * host;
  const char * user;
  const char * pass;
  const char * dbase;
  char * info;
  int anz_db  = 0;
  int anz_tab = 0;

  columnTypes = NULL;

  connectionEnabled=false;

  QObject * myApp = parent();
  QApplication::setOverrideCursor( Qt::WaitCursor );
  
  if ( ! url.isEmpty() ){
    
    // std::cout << "Aufruf von openDocument" << endl;
    
    host = url.latin1();
    user = uname.latin1();
    pass = pwd.latin1();
    dbase = base.latin1();
    char*  port = "00000000";
    
    char * portpart = strchr(host, ':');
    if(portpart != NULL){
      doc_port = atoi(portpart+1);
      sprintf( port, "%d", doc_port);

      // // std::cout << port << endl;

      *portpart = 0;
      // std::cout << "host: " << host << endl << "port: " << doc_port << endl;
    }else{
      // std::cout << "no port: " << host << endl;
      port = "";
    }
    
    doc_host.setLatin1(host);
    doc_uname.setLatin1(user);
    doc_pass.setLatin1(pass);
    doc_base.setLatin1(dbase);

    QString * cinfo = new QString("host='");
    cinfo->append( host );
    cinfo->append("' user='");
    cinfo->append(user);
    cinfo->append("' port='");
    cinfo->append(port);
    cinfo->append("' dbname='");
    cinfo->append(dbase);
    cinfo->append("' password='");
    cinfo->append(pass);
    cinfo->append("'");


    unsigned int num_rows;
    unsigned int i;
    
    PGconn     *conn;
    PGresult   *res;
    conn = PQconnectdb( cinfo->latin1() );
    if ( PQstatus(conn) == CONNECTION_BAD ){
      // std::cout << "no conn! ERROR connecting to server... :-((" << endl;
      // std::cout << PQerrorMessage(conn) << endl;
      QString *errmsg = new QString( kl->translate( i18n("Error connecting to server.")));
      QApplication::restoreOverrideCursor();
      
      PQfinish(conn);
      
      doc_host.setLatin1("");
      doc_uname.setLatin1("");
      doc_pass.setLatin1("");
      
      emit( DbQueryError( errmsg) );
      delete DB;
      DB=NULL;
      QApplication::restoreOverrideCursor();
      return false;
    } else {

      /*********************************** Look for Databases *******************************************************/

      clearResult();
      DB=NULL;
      base_t * hilf=NULL;

       /* start a transaction block */
       res = PQexec(conn, "BEGIN");
       if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
         {
           // std::cout << "no begin! ERROR connecting to server... :-((" << endl;
           PQclear(res);
           PQfinish(conn);
           QString *errmsg = new QString( kl->translate( i18n("Error connecting to server.")));
           emit( DbQueryError( errmsg) );
           QApplication::restoreOverrideCursor();
           return false;
         }
      
      PQclear(res);
      
      res = PQexec(conn, "DECLARE mycursor CURSOR FOR SELECT datname FROM pg_database");
      if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
        {
          // std::cout << "ERROR connecting to server... :-((" << endl;
          // std::cout << PQerrorMessage(conn) << endl;

          PQclear(res);
          PQfinish(conn);
          QString *errmsg = new QString( kl->translate( i18n("Error connecting to server.")));
          emit( DbQueryError( errmsg) );
          QApplication::restoreOverrideCursor();
          return false;
        }

      PQclear(res);
      res = PQexec(conn, "FETCH ALL in mycursor");
      if (!res || PQresultStatus(res) != PGRES_TUPLES_OK)
        {
          // std::cout << "no fetchall! ERROR connecting to server... :-((" << endl;
          PQclear(res);
          PQfinish(conn);
          QString *errmsg = new QString( kl->translate( i18n("Error connecting to server.")));
          emit( DbQueryError( errmsg) );
          QApplication::restoreOverrideCursor();
          return false;       
        }
                
      int nFields = PQnfields(res);

      for (i = 0; i < PQntuples(res); i++)
        {
          base_t * temp = new base_t;
          char * db_name=NULL;
          db_name = new char[ strlen( PQgetvalue(res, i, 0) ) + 1 ];
          db_name = strcpy( db_name , PQgetvalue(res, i, 0) );
          temp->name = db_name;
          temp->table_list = NULL;
          if (DB==NULL){
            DB = temp;
          } else {
            hilf->next = temp;
          }
          hilf = temp;
        }
      hilf->next=NULL;
      PQclear(res);
        
      res = PQexec(conn, "CLOSE mycursor");
      PQclear(res);



      /*****************************************************************************************************************************/
      // Hier alle Typen-Namen und OIDs auslesen und columnTypes setzen:
      // type = oid
      // name = name
      res = PQexec(conn, "DECLARE mycursor CURSOR FOR SELECT oid,typname FROM pg_type");
      if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
        {
          // std::cout << "ERROR connecting to server... :-((" << endl;
          // std::cout << PQerrorMessage(conn) << endl;

          PQclear(res);
          PQfinish(conn);
          QString *errmsg = new QString( kl->translate( i18n("Error connecting to server.")));
          emit( DbQueryError( errmsg) );
        }

      PQclear(res);
      res = PQexec(conn, "FETCH ALL in mycursor");
      if (!res || PQresultStatus(res) != PGRES_TUPLES_OK)
        {
          // std::cout << "no fetchall! ERROR connecting to server... :-((" << endl;
          PQclear(res);
          PQfinish(conn);
          QString *errmsg = new QString( kl->translate( i18n("Error connecting to server.")));
          emit( DbQueryError( errmsg) );
        }
                
      nFields = PQnfields(res);
      column_t* hilfcol;
      for (i = 0; i < PQntuples(res); i++)
        {
          column_t * temp = new column_t;
          char * col_n=NULL;
          char * col_o=NULL;
          col_n = new char[ strlen( PQgetvalue(res, i, 1) ) + 1 ];
          col_n = strcpy( col_n , PQgetvalue(res, i, 1) );
          col_o = new char[ strlen( PQgetvalue(res, i, 0) ) + 1 ];
          col_o = strcpy( col_o , PQgetvalue(res, i, 0) );

          temp->name = col_n;
          temp->type = col_o;

          // std::cout << "Name: " << col_n << " OID: " << col_o << endl;
          if (columnTypes==NULL){
            columnTypes = temp;
          } else {
            hilfcol->next = temp;
          }
          hilfcol = temp;
        }
      hilfcol->next=NULL;
      PQclear(res);
        
      res = PQexec(conn, "CLOSE mycursor");
      PQclear(res);

      // bis hier


      res = PQexec(conn, "COMMIT");
      PQclear(res);
        
      PQfinish(conn);
      
      connectionEnabled = true;  
    }

    base_t  *dats = DB;

    while ( dats ){

      /*********************************** Try to connect to all dbs ************************************************/

      cinfo->setLatin1("host='");
      cinfo->append( host );
      cinfo->append("' user='");
      cinfo->append(user);
      cinfo->append("' port='");
      cinfo->append(port);
      cinfo->append("' dbname='");
      cinfo->append( dats->name );
      cinfo->append("' password='");
      cinfo->append(pass);
      cinfo->append("'");

      /*********************************** Look for Tables **********************************************************/

      QString *query = new QString("");
      query->append( "DECLARE mycursor CURSOR FOR SELECT relname FROM pg_class WHERE relkind like 'r'" );
      
      table_t *tbs  = NULL;

      conn = PQconnectdb( cinfo->latin1() );
      if ( PQstatus(conn) == CONNECTION_BAD ){
        // std::cout << "can't read tables for " << dats->name << " ... :-((" << endl;
        // std::cout << PQerrorMessage(conn) << endl;
        // try next base:
        dats = dats->next;
        PQfinish(conn);
        
      } else {      
        // Start transaction-block
        res = PQexec(conn, "BEGIN");
        if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
          {
            // std::cout << "no begin! ERROR connecting to server... :-((" << endl;
            QString *errmsg = new QString( PQerrorMessage(conn) );
            emit( DbQueryError( errmsg) );
          }
        
        // send query
        PQclear(res);
        res = PQexec(conn, query->latin1() );
        if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
          {
            // std::cout << "ERROR connecting to server... :-((" << endl;
            // std::cout << PQerrorMessage(conn) << endl;
            QString *errmsg = new QString( PQerrorMessage(conn) );
            emit( DbQueryError( errmsg) );
          }
        
        PQclear(res);
        res = PQexec(conn, "FETCH ALL in mycursor");
        if (!res || PQresultStatus(res) != PGRES_TUPLES_OK)
          {
            // std::cout << "no fetchall! ERROR connecting to server... :-((" << endl;
            QString *errmsg = new QString( PQerrorMessage(conn) );
            emit( DbQueryError( errmsg) );
          }
        
        int nFields = PQnfields(res);
        
        for (i = 0; i < PQntuples(res); i++)
          {
            table_t * temp = new table_t;
            char * tb_name=NULL;
            tb_name = new char[ strlen( PQgetvalue(res, i, 0) ) + 1 ];
            tb_name = strcpy( tb_name , PQgetvalue(res, i, 0) );
            temp->name = tb_name;
            temp->column_list = NULL;
            temp->type = 0;
            temp->type_n = " ";
            if (dats->table_list==NULL){
              dats->table_list = temp;
            } else {
              tbs->next = temp;
            }
            tbs = temp;
          }
        tbs->next=NULL;
        PQclear(res);
        
        res = PQexec(conn, "CLOSE mycursor");
        PQclear(res);
        
        res = PQexec(conn, "COMMIT");
        PQclear(res);
        dats = dats->next;
        PQfinish(conn);      
      }

    }

    // We have bases and tables, so lets look for the columns:
    dats = DB;

    while ( dats ){
      
      /*********************************** Try to connect to all dbs ************************************************/


      cinfo->setLatin1("host='");
      cinfo->append( host );
      cinfo->append("' user='");
      cinfo->append(user);
      cinfo->append("' port='");
      cinfo->append(port);
      cinfo->append("' dbname='");
      cinfo->append( dats->name );
      cinfo->append("' password='");
      cinfo->append(pass);
      cinfo->append("'");

      conn = PQconnectdb( cinfo->latin1() );
      if ( PQstatus(conn) == CONNECTION_BAD ){// if connstatus bad
        // std::cout << "can't read tables for " << dats->name << " ... :-((" << endl;
        // std::cout << PQerrorMessage(conn) << endl;
        
      } else { //if constatus bad else

        /*********************************** Look for Column-Names *******************************************************/

        // Start transaction-block
        res = PQexec(conn, "BEGIN");
        if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
          {
            // std::cout << "no begin! ERROR connecting to server... :-((" << endl;
          }
        
        table_t* tbs = dats->table_list;
        while( tbs ){

          QString *query = new QString("");
          query->append( "DECLARE mycursor CURSOR FOR SELECT pg_attribute.attname,pg_class.oid,pg_type.typname FROM pg_attribute,pg_class,pg_type WHERE pg_attribute.attrelid = pg_class.oid and pg_class.relname like '" );
          query->append( tbs->name);
          query->append( "' and pg_attribute.attnum>0 and pg_attribute.atttypid=pg_type.oid");

          PQclear(res);
          res = PQexec(conn, query->latin1() );
          if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
            {
              // std::cout << "ERROR connecting to server... :-((" << endl;
              // std::cout << PQerrorMessage(conn) << endl;
              QString *errmsg = new QString( PQerrorMessage(conn) );
              emit( DbQueryError( errmsg) );
            }
          
          PQclear(res);
          res = PQexec(conn, "FETCH ALL in mycursor");
          if (!res || PQresultStatus(res) != PGRES_TUPLES_OK)
            {
              // std::cout << "no fetchall! ERROR connecting to server... :-((" << endl;
              QString *errmsg = new QString( PQerrorMessage(conn) );
              emit( DbQueryError( errmsg) );
            }

          char * tab_oid;
          
          /* get results */
          int nFields = PQnfields(res);
          column_t* cls = NULL;
          for (i = 0; i < PQntuples(res); i++)
            {
              column_t * temp = new column_t;
              char * cl_name=NULL;
              char * cl_type=NULL;
              char cl_pri = ' ';
              char * cl_extra=NULL;
              

              cl_name = new char[ strlen( PQgetvalue(res, i, 0) ) + 1 ];
              cl_name = strcpy( cl_name , PQgetvalue(res, i, 0) );

              tab_oid = new char[ strlen( PQgetvalue(res, i, 1) ) + 1 ];
              tab_oid = strcpy( tab_oid , PQgetvalue(res, i, 1) );

              cl_type = new char[ strlen( PQgetvalue(res, i, 2) ) + 1 ];
              cl_type = strcpy( cl_type , PQgetvalue(res, i, 2) );

              cl_extra = new char[ strlen( "" ) + 1 ];
              cl_extra = strcpy( cl_extra , "" );

              temp->name = cl_name;
              temp->type = cl_type;
              temp->extra_flags = cl_extra;
              temp->pri = cl_pri;

              
              if (tbs->column_list==NULL){
                if (tbs) tbs->column_list = temp;
              } else {
                if (cls) cls->next = temp;
              }
              cls = temp;
            } // ende for PQntupels
          if (cls) cls->next=NULL;
          PQclear(res);
          
          /* close the cursor */
          res = PQexec(conn, "CLOSE mycursor");

          /*********************************** Get Default-Values *****************************************************/
          query->setLatin1( "DECLARE mycursor CURSOR FOR SELECT pg_attribute.attname,pg_class.oid,pg_attrdef.adsrc FROM pg_attribute,pg_class,pg_attrdef WHERE pg_attribute.attrelid = pg_class.oid and pg_class.relname like '" );
          query->append( tbs->name);
          query->append( "' and pg_class.oid=pg_attrdef.adrelid and pg_attribute.attnum=pg_attrdef.adnum");

//          // std::cout << query->latin1() << endl;

          PQclear(res);
          res = PQexec(conn, query->latin1() );
          if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
            {
              // std::cout << "ERROR connecting to server... :-((" << endl;
              // std::cout << PQerrorMessage(conn) << endl;
              QString *errmsg = new QString( PQerrorMessage(conn) );
              emit( DbQueryError( errmsg) );
            }
          
          PQclear(res);
          res = PQexec(conn, "FETCH ALL in mycursor");
          if (!res || PQresultStatus(res) != PGRES_TUPLES_OK)
            {
              // std::cout << "no fetchall! ERROR connecting to server... :-((" << endl;
              QString *errmsg = new QString( PQerrorMessage(conn) );
              emit( DbQueryError( errmsg) );
            }

          /* get results */
          nFields = PQnfields(res);
          for (i = 0; i < PQntuples(res); i++)
            {

              QString* def;
              QString* co;
              co = new QString( PQgetvalue(res, i, 0) );
              def = new QString( PQgetvalue(res, i, 2) );

              column_t* c = tbs->column_list;
              while(c){
                if( co->compare( c->name ) == 0 ){
                  if( def->contains("nextval",false) ){
                    c->extra_flags = "auto_increment";
                  }
                }
                c = c->next;
              }
              
            } // ende for PQntupels
          PQclear(res);
          
          /* close the cursor */
          res = PQexec(conn, "CLOSE mycursor");

          
          /*********************************** Look for Indexes *******************************************************/
          //  // std::cout << "Table-OID: " << tab_oid << endl;

          query->setLatin1( "DECLARE mycursor CURSOR FOR SELECT indkey FROM pg_index WHERE indrelid LIKE '" );
          query->append( tab_oid );
          query->append( "'" );
          //  query->append( "' and indisprimary = true" );
          
          PQclear(res);
          res = PQexec(conn, query->latin1() );
          if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
            {
              // std::cout << "ERROR connecting to server... :-((" << endl;
              QString *errmsg = new QString( PQerrorMessage(conn) );
              emit( DbQueryError( errmsg) );
            }
          
          PQclear(res);
          res = PQexec(conn, "FETCH ALL in mycursor");
          if (!res || PQresultStatus(res) != PGRES_TUPLES_OK)
            {
              // std::cout << "no fetchall! ERROR connecting to server... :-((" << endl;
              QString *errmsg = new QString( PQerrorMessage(conn) );
              emit( DbQueryError( errmsg) );
            }
          
          // Place for 100 keys (do we need more???)
          char * tab_keys[100];
          int j = 0;
          int num_keys;
          /* get results */
          nFields = PQnfields(res);

          for (i = 0; i < PQntuples(res); i++){
            
             QString * key = new QString( PQgetvalue(res, i, 0) ); 
             int c_num = key->contains(' ');
             if(c_num){
               for(int d=0; d<c_num; d++){
                 QString keycolname = key->section(' ',d,d);
                 tab_keys[j] = new char[ strlen( keycolname.latin1() ) + 1 ];
                 strcpy( tab_keys[j] , keycolname.latin1() );
                 j++;
               }
             } else {
               tab_keys[j] = new char[ strlen( PQgetvalue(res, i, 0) ) + 1 ];
               tab_keys[j] = strcpy ( tab_keys[j] , PQgetvalue(res, i, 0) );
               j++;
             }

          } // ende for PQntupels
          num_keys = j;
          PQclear(res);
          res = PQexec(conn, "CLOSE mycursor");

          /*************************************************************************************************************/

          /********************************* Now get column-names for indexes ******************************************/

          for ( int d = 0; d < num_keys; d++){
            if ( tab_keys[d] > 0){
              //  // std::cout << "Column-Numbers: " << tab_keys[d] << endl;
              query->setLatin1("DECLARE mycursor CURSOR FOR SELECT attname FROM pg_attribute WHERE attrelid LIKE '" );
              query->append( tab_oid );
              query->append( "' AND attnum LIKE '" );
              query->append( tab_keys[d] );
              query->append( "'");

              PQclear(res);
              res = PQexec(conn, query->latin1() );
              if (!res || PQresultStatus(res) != PGRES_COMMAND_OK)
                {
                  // std::cout << "ERROR connecting to server... :-((" << endl;
                  QString *errmsg = new QString( PQerrorMessage(conn) );
                  emit( DbQueryError( errmsg) );
                }
              
              PQclear(res);
              res = PQexec(conn, "FETCH ALL in mycursor");
              if (!res || PQresultStatus(res) != PGRES_TUPLES_OK)
                {
                  // std::cout << "no fetchall! ERROR connecting to server... :-((" << endl;
                  QString *errmsg = new QString( PQerrorMessage(conn) );
                  emit( DbQueryError( errmsg) );
                }
              
              // Place for 100 keys (do we need more???)
              char * tab_keys[100];
              int j = 0;
              int num_keys;
              /* get results */
              nFields = PQnfields(res);
              
              for (i = 0; i < PQntuples(res); i++){
                
                QString * indexColumnName = new QString( PQgetvalue(res, i, 0) ); 
                column_t * columnlist = tbs->column_list;
                while( columnlist ){
                  if ( indexColumnName->compare( columnlist->name ) == 0 ){
                    columnlist->pri = 'Y';
                  }
                  columnlist = columnlist->next;
                }
                
              } // ende for PQntupels
              num_keys = j;
              PQclear(res);
              res = PQexec(conn, "CLOSE mycursor");
            }
          }
          
          /*************************************************************************************************************/


          tbs = tbs->next; 
        } // ende while tbs

        /* commit the transaction */
        PQclear(res);
        res = PQexec(conn, "COMMIT");
        PQclear(res);

      } // ende if connection-bad else 

      dats = dats->next;
      PQfinish(conn);      

    } // ende while dats

  }   // Ende if url


  emit( connectionChanged() );
  
  if ( RESULT ){
    delete RESULT;
    RESULT = NULL;
  }
  if ( RESULT_Start ){
    delete RESULT_Start;
    RESULT_Start = NULL;
  }
  
  if(DB){
    basePosition = DB;
    tablePosition = DB->table_list;
    if (tablePosition) {
      columnPosition = tablePosition->column_list;
    }else{
      columnPosition = NULL;
    }
    rowPosition = NULL;      // not needed (implemented) now
    fieldPosition = NULL;    // not needed (implemented) now
  } else {
    basePosition = NULL;
    tablePosition = NULL;
    columnPosition = NULL;
    rowPosition = NULL;      // not needed (implemented) now
    fieldPosition = NULL;    // not needed (implemented) now
  }
  
  // emit( resultChanged() );
  emit loginComplete();
  connectionEnabled=true;
  if (! DB ){
    connectionEnabled=false;
    emit loginError();
  }
  QApplication::restoreOverrideCursor();
  return true;
}


extern "C" SqlGuiDoc* create(QWidget *parent, const char *name){
  return new SqlGuiDocPgsql( parent, name );
}

extern "C" void destroy( SqlGuiDoc* p){
  delete p;
}

//#include "sqlguidoc_pgsql.moc.cpp"
